import importlib

import numpy as np
import pytest

from pybest import filemanager

ent = importlib.import_module(".", "pybest-entanglement")


def test_read_s1_data():
    "Write a file, read it, and compare with expected output."
    numbers = """1 0.1\n2 0.2\n3 0.3"""
    tmpfile = filemanager.temp_path("s1_test.dat")
    with open(tmpfile, "w+") as data:
        data.write(numbers)
    index, value = ent.read_s1_data(1, None, tmpfile)
    assert np.allclose(index, np.asarray([1, 2, 3]))
    assert np.allclose(value, np.asarray([0.1, 0.2, 0.3]))


def test_read_i12_data():
    "Write a file, read it, and compare with expected output."
    numbers = """1 2 0.1\n1 3 0.2\n2 3 0.3"""
    tmpfile = filemanager.temp_path("i12_test.dat")
    with open(tmpfile, "w+") as data:
        data.write(numbers)
    index1, index2, value = ent.read_i12_data(1, 3, tmpfile)
    assert np.allclose(index1, np.asarray([1, 1, 2]))
    assert np.allclose(index2, np.asarray([2, 3, 3]))
    assert np.allclose(value, np.asarray([0.1, 0.2, 0.3]))


def test_get_highest_s1index():
    "Check if a function returns correct indices."
    index = [1, 2, 3, 4, 5, 6, 7, 8, 9]
    value = [1, 0.5, 0, 2, 0, 3, 0, 0, 1.1]
    out = ent.get_highest_s1_index(index, value, 4)
    assert out == [1, 4, 6, 9]


def test_get_highest_i12index():
    "Check if a function returns correct indices."
    index1 = [1, 1, 1, 1, 2, 2, 2, 3, 3, 4]
    index2 = [2, 3, 4, 5, 3, 4, 5, 4, 5, 5]
    value = [1, 0, 0, 2, 0, 0, 1, 0, 0, 0.25]
    out = ent.get_highest_i12_index(index1, index2, value, 4)
    assert out == [1, 2, 4, 5]


default_parameters = {
    "f": None,
    "i": 1,
    "iname": "i12.dat",
    "indices": None,
    "largest_i": 0,
    "largest_s": 0,
    "order": None,
    "plot": "all",
    "sname": "s1.dat",
    "threshold": 0.001,
    "zoom": 0.03,
}

test_parse_args = [
    {
        "input": [],
    },
    {
        "input": ["-i", "2", "-f", "5", "--sname", "s.dat"],
        "f": 5,
        "i": 2,
        "sname": "s.dat",
    },
    {
        "input": ["--zoom", "0.125", "--threshold", "0.0625"],
        "threshold": 0.0625,
        "zoom": 0.125,
    },
    {
        "input": ["--largest-i", "9", "--iname", "i.dat"],
        "iname": "i.dat",
        "largest_i": 9,
    },
    {
        "input": ["--largest-s", "3", "--order", "3", "2", "1"],
        "largest_s": 3,
        "order": [3, 2, 1],
    },
    {
        "input": ["--indices", "2", "1", "3", "7", "--plot", "i"],
        "indices": [2, 1, 3, 7],
        "plot": "i",
    },
]


@pytest.mark.parametrize("test_case", test_parse_args)
def test_parse_args(test_case):
    """Test parsing args."""
    parser = ent.parse_args(test_case["input"])
    all_parameter_names = default_parameters.keys()
    for param in all_parameter_names:
        value = getattr(parser, param)
        if param in test_case:
            # Check args provided directly
            assert value == test_case[param], f"incorrect {param}"
        else:
            # Check default args
            assert value == default_parameters[param], f"incorrect {param}"


def test_ploti12():
    "Test passes if calling a function ploti12 does not raise an error."
    ent.ploti12(
        np.asarray([1, 1, 1, 1, 2, 2, 2, 3, 3, 4]),
        np.asarray([2, 3, 4, 5, 3, 4, 5, 4, 5, 5]),
        np.asarray([1, 0, 0, 1, 0, 0, 1, 0, 0, 0.25]),
        np.asarray([1, 2, 3, 4, 5]),
        0.0001,
        np.asarray([1, 2, 3, 4, 5]),
        np.asarray([0.1, 0.2, 0.3, 0.2, 0.1]),
        0.25,
    )


def test_plots1():
    "Test passes if calling a function plots1 does not raise an error."
    ent.plots1(1, 4, np.asarray([1, 2, 3]), np.asarray([0.1, 0.2, 0.3]), 0.4)
