#!/usr/bin/env python3
# PyBEST: Pythonic Black-box Electronic Structure Tool
# Copyright (C) 2016-- The PyBEST Development Team
#
# This file is part of PyBEST.
#
# PyBEST is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# PyBEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
# --
import os
from glob import glob

from setuptools import find_packages
from setuptools_scm import get_version

# NOTE: when skbuild transitions to scikit-build-core
# it will support editable builds by a default
# which should be installable by simple `python3 -m pip install -e .`
# however, it does not, so for now `python3 setup.py develop` must suffice
# NOTE: https://github.com/scikit-build/scikit-build/issues/981
# NOTE: https://github.com/scikit-build/scikit-build/issues/1025
from skbuild import setup

# populate install requires
# -------------------------
with open("requirements.txt", encoding="utf-8") as f:
    required = f.read().splitlines()

# populate long description
# -------------------------
with open("README.md", encoding="utf-8") as f:
    long_description = f.read()


# NOTE: python3 -m auditwheel repair dist/*.whl --exclude libmkl_rt.so.1  <- we need to parse lddtree, to exclude libmkl correctly
# TODO: remove it when transition to scikit-build-core happens
def _generate_dynamic_cmake_args():
    """Small wrapper to choose correct build options,
    and hint location of dependencies
    """
    args = []
    if os.environ.get("LIBINT2_ROOT"):
        args.append(f"-DPYBEST_LIBINT2_ROOT={os.environ['LIBINT2_ROOT']}")
    if os.environ.get("LIBCHOL_ROOT"):
        args.append("-DPYBEST_ENABLE_CHOLESKY=ON")
        args.append(f"-DPYBEST_LIBCHOL_ROOT={os.environ['LIBCHOL_ROOT']}")
    # NOTE: we either use MKL, or default to OpenBLAS
    if os.environ.get("PYBEST_USE_MKL"):
        args.append("-DUSE_MKL:BOOL=ON")
    else:
        args.append("-DUSE_OPENBLAS:BOOL=ON")
    return args


setup(
    name="pybest",
    version=get_version(),
    description="PyBEST: Pythonic Black-box Electronic Structure Tool.",
    author="Somayeh Ahmadkhani <so.ahmadkhani@gmail.com>, "
    "Katharina Boguslawski <k.boguslawski@fizyka.umk.pl>, "
    "Dariusz Kędziera <teodar@chem.umk.pl>, "
    "Maximilian Kriebel <maximilian.kriebel@web.de>, "
    "Paweł Tecmer <ptecmer@fizyka.umk.pl>, "
    "Piotr Żuchowski <pzuch@fizyka.umk.pl>, "
    "Saman Behjou <s.behjoo@gmail.com>, "
    "Filip Brzęk <filip.brzek@gmail.com>, "
    "Kacper Cieślak <kacperfonex@gmail.com>, "
    "Rahul Chakraborty <rahuldev.chakraborty91@gmail.com>, "
    "Antonina Dobrowolska <antosiadob@gmail.com>, "
    "Seyedehdelaram Jahani <aramehjahani11@gmail.com>, "
    "Zahra Karimi <karimizahra369@gmail.com>, "
    "Michał Kopczyński <michal.kopczynski3.14@gmail.com>, "
    "Aleksandra Leszczyk <aleksandra.leszczyk@gmail.com>, "
    "Artur Nowak <nowak.art90@gmail.com>, "
    "Lena Szczuczko <302741@stud.umk.pl>, "
    "Emil Sujkowski <e.sujkowski@gmail.com>, "
    "Julian Świerczyński <icytwv@gmail.com>",
    url="http://www.fizyka.umk.pl/~pybest",
    scripts=glob("scripts/*.py"),
    package_dir={"": "src"},
    packages=find_packages("src"),
    include_package_data=True,
    package_data={
        "pybest": [
            "data/basis/*.g94",
            "data/elements.csv",
            "data/*.bib",
            "data/examples/*/*.py",
            "data/test/*.*",
            "core/*.h",
            "core/CMakeLists.txt",
        ],
    },
    cmake_source_dir=os.path.join("src", "pybest", "core"),
    cmake_install_dir=os.path.join("src", "pybest", "core"),
    cmake_args=[
        "-DCMAKE_BUILD_TYPE=Release",
        "-DCMAKE_VERBOSE_MAKEFILE:BOOL=ON",
        *_generate_dynamic_cmake_args(),
    ],
    install_requires=required,
    python_requires=">=3.9,<4.0",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Operating System :: MacOS",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3",
        "Programming Language :: C++",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Scientific/Engineering :: Chemistry",
    ],
    zip_safe=False,
)
