// PyBEST: Pythonic Black-box Electronic Structure Tool
// Copyright (C) 2016-- The PyBEST Development Team
//
// This file is part of PyBEST.
//
// PyBEST is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// PyBEST is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <http://www.gnu.org/licenses/>
// --

#include <nanobind/nanobind.h>
#include <nanobind/stl/string.h>
#include <nanobind/stl/vector.h>

#include "basis.h"
#include "external_charges.h"
#include "fourindex.h"
#include "ints.h"
#include "libint_utils.h"
#include "sapt_utils.h"
#include "static_embedding.h"

namespace nb = nanobind;

NB_MODULE(core, m) {
  m.doc() =
      "PyBEST core module aggregating various sub-modules ";  // optional module
                                                              // docstring

  auto m_transform = m.def_submodule(
      "transform", "PyTransform module with explicit transformation schemes");
  auto m_ints = m.def_submodule(
      "ints", "module PyInts to calculate various integrals using libint2");
  auto m_sapt_core =
      m.def_submodule("sapt_core", "sapt core module with sapt utils routines");

  auto m_basis = m.def_submodule(
      "basis", "nanobind module basis to read AO basis sets using libint2 API");

  // py_transform
  m_transform.def("apply_four_index_transform_cpp",
                  &py_transform::apply_four_index_transform_cpp,
                  "Perform 4-index transform of dense object",
                  nb::rv_policy::take_ownership);
  m_transform.def("apply_four_index_transform_cpp_chol",
                  &py_transform::apply_four_index_transform_cpp_chol,
                  "Perform 4-index transform of Cholesky object",
                  nb::rv_policy::take_ownership);

  // py_ints
  m_ints.def(
      "compute_nuclear_repulsion_cpp", &py_ints::compute_nuclear_repulsion,
      "A function to compute the potential energy due to nuclear-nuclear "
      "repulsion");
  m_ints.def(
      "compute_external_charges_cpp", &py_ints::compute_external_charges,
      "A function to compute the potential energy due to the interaction "
      " between point charges");
  m_ints.def(
      "compute_nuclear_pc_cpp", &py_ints::compute_nuclear_pc,
      "A function to compute the potential energy due to the interaction of"
      " the external point charges and the nuclei");
  m_ints.def("compute_overlap_cpp", &py_ints::compute_overlap,
             "A function to compute the AO overlap",
             nb::rv_policy::take_ownership);
  m_ints.def(
      "compute_kinetic_cpp", &py_ints::compute_kinetic,
      "A function to compute the kinetic energy integrals in the AO basis",
      nb::rv_policy::take_ownership);
  m_ints.def(
      "compute_nuclear_cpp", &py_ints::compute_nuclear,
      "A function to compute the electron-nuclei attraction integrals in the "
      "AO basis",
      nb::rv_policy::take_ownership);
  m_ints.def(
      "compute_point_charges_cpp", &py_ints::compute_point_charges,
      "A function that computes the interaction between external point charges "
      "and electrons",
      nb::rv_policy::take_ownership);
#ifdef PYBEST_ENABLE_PVP
  m_ints.def(
      "compute_pvp_cpp", &py_ints::compute_pvp,
      "A function to compute the pVp integrals in the uncontracted AO basis",
      nb::rv_policy::take_ownership);
  m_ints.def(
      "compute_ppcp_cpp", &py_ints::compute_ppcp,
      "A function to compute the pPCp integrals in the uncontracted AO basis",
      nb::rv_policy::take_ownership);
#endif
  m_ints.def("compute_dipole_cpp", &py_ints::compute_dipole,
             "A function to compute the dipole integrals in the AO basis",
             nb::rv_policy::take_ownership);
  m_ints.def("compute_quadrupole_cpp", &py_ints::compute_quadrupole,
             "A function to compute the quadrupole integrals in the AO basis",
             nb::rv_policy::take_ownership);
  m_ints.def(
      "compute_eri_cpp", &py_ints::compute_eri,
      "A function to compute the electron repulsion integrals in the AO basis",
      nb::rv_policy::take_ownership);
#ifdef PYBEST_ENABLE_CHOLESKY
  m_ints.def("compute_cholesky_eri_cpp", &py_ints::compute_cholesky_eri,
             "A function to compute Cholesky decomposed electron repulsion "
             "integrals in the AO basis",
             nb::rv_policy::take_ownership);
#endif

  // py_sapt_core
  m_sapt_core.def(
      "get_amplitudes", &py_sapt_core::get_sapt_amplitudes,
      "Returns SAPT uncoupled restricted theory denominators used in "
      "amplitudes");

  // py_basis utility functions
  m_basis.def("get_nubf", &py_libint_utils::get_nubf,
              "Return number of uncontracted basis functions");
  m_basis.def("get_solid_tform_u2c", &py_libint_utils::get_solid_tform_u2c,
              "Return the transformation matrix uncontracted<-->contracted",
              nb::rv_policy::take_ownership);
  // py_basis Basis constructor from python-side
  m_basis.def("from_coordinates", &py_libint_utils::from_coordinates,
              "Read basis set from coordinates");

  // py_basis Basis class export, it's exported to core directly, to allow
  // direct imports, and efficient monkey-patching within gbasis/__init__.py
  nb::class_<Basis> basis(m, "Basis");
  basis
      .def(nb::init<const std::string &, const std::string &,
                    const std::string &>())
      .def(nb::init<const std::string &, std::vector<long>, std::vector<long>,
                    std::vector<int>, std::vector<double>,
                    std::vector<double>>())
      .def(nb::init<const Basis &>())
      .def("print_basis_info", &Basis::print_basis_info)
      .def("print_atomic_info", &Basis::print_atomic_info)
      .def("renormalize_contr", &Basis::renormalize_contr)
      .def("get_nbasis_in_shell", &Basis::get_nbasis_in_shell)
      .def("set_dummy_atoms", &Basis::set_dummy_atom)
      .def("dump_molden", &Basis::dump_molden)
      .def("dump_cube_orbital", &Basis::dump_cube_orbital)
      // FIXME: test for negative values
      .def("get_shell_size", &Basis::get_shell_size)
      .def("get_renorm_pure",
           (double(Basis::*)(std::size_t, std::size_t, std::size_t)) &
               Basis::get_renorm,
           "Get renormalization constant for solid harmonics")
      .def("get_renorm_cart",
           (double(Basis::*)(std::size_t, std::size_t, std::size_t,
                             std::vector<std::size_t>)) &
               Basis::get_renorm,
           "Get renormalization constant for cartesians")
      // getters; we do not want them to be writable from python
      .def_prop_ro("basisname", &Basis::getBasisname, nullptr,
                   "Get the name of the AO basis set (if specified)")
      .def_prop_ro("molfile", &Basis::getMolfile, nullptr,
                   "Get the name of the coordinate xyz file (if present)")
      .def_prop_ro("basisdir", &Basis::getDirname, nullptr,
                   "Get the directory of basis set library (if present)")
      // define read-writable attributes
      .def_rw("nbasis", &Basis::nbasis, "the total number of basis functions")
      .def_rw("ncenter", &Basis::ncenter, "the total number of atomic centers")
      .def_rw("nshell", &Basis::nshell, "the total number of shells")
      .def_rw("nprim", &Basis::nprim,
              "the number of primitives per contraction")
      .def_rw("alpha", &Basis::alpha, "the exponents for each primitive")
      .def_rw("contraction", &Basis::contraction,
              "the contraction coefficients for each primitive")
      .def_rw("shell2atom", &Basis::shell2atom, "the shell to atom list")
      .def_rw("shell_types", &Basis::shell_types,
              "the shell type per contraction")
      .def_rw("atom", &Basis::atomic_numbers, "the atomic numbers of each atom")
      .def_rw("coordinates", &Basis::coordinates, "Get the coordinates");

  // StaticEmbedding class
  nb::class_<StaticEmbedding> staticembedding(m, "StaticEmbedding");
  staticembedding
      .def(nb::init<std::vector<double>, std::vector<double>,
                    std::vector<double>, std::vector<double>,
                    std::vector<double>>())
      .def(nb::init<const StaticEmbedding &>())
      .def("compute_static_embedding_cpp",
           &StaticEmbedding::compute_static_embedding,
           "A function to calculate static"
           "embedding taken from an external file",
           nb::rv_policy::take_ownership);

  // ExternalCharges class
  nb::class_<ExternalCharges> externalcharges(m, "ExternalCharges");
  externalcharges
      .def(nb::init<std::vector<double>, std::vector<double>,
                    std::vector<double>, std::vector<double>>())
      .def(nb::init<const ExternalCharges &>())
      .def_rw("charges", &ExternalCharges::charges, "charge values")
      .def_rw("coordinates", &ExternalCharges::coordinates);
}
