# PyBEST: Pythonic Black-box Electronic Structure Tool
# Copyright (C) 2016-- The PyBEST Development Team
#
# This file is part of PyBEST.
#
# PyBEST is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# PyBEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
# --
"""PyBest core module aggregating various C-extension sub-modules"""

from __future__ import annotations

import typing

import h5py._hl.group
import numpy as np
from numpy.typing import NDArray

class Basis:
    @staticmethod
    def from_hdf5(group: h5py._hl.group.Group) -> Basis:
        """Method used as a class method, in monkeypatched core.Basis (nanobind export)
        return an instance of core.Basis object from hdf5 dump.

        Args:
            group (h5py.Group): group in which Basis object information is stored

        Returns:
            Basis: core.Basis object (nanobind export)

        """

    @typing.overload
    def __init__(self, arg0: str, arg1: str, arg2: str) -> None: ...
    @typing.overload
    def __init__(
        self,
        arg0: str,
        arg1: list[int],
        arg2: list[int],
        arg3: list[int],
        arg4: list[float],
        arg5: list[float],
    ) -> None: ...
    @typing.overload
    def __init__(self, arg0: Basis) -> None: ...
    def dump_cube_orbital(
        self,
        arg0: str,
        arg1: NDArray[np.float64],
        arg2: NDArray[np.float64],
        arg3: NDArray[np.float64],
        arg4: NDArray[np.float64],
    ) -> None: ...
    def dump_molden(
        self,
        arg0: str,
        arg1: NDArray[np.float64],
        arg2: NDArray[np.float64],
        arg3: NDArray[np.float64],
        arg4: NDArray[np.bool_],
        arg5: float,
    ) -> None: ...
    def get_nbasis_in_shell(self, arg0: int) -> int: ...
    def get_renorm_cart(
        self, arg0: int, arg1: int, arg2: int, arg3: list[int]
    ) -> float:
        """Get renormalization constant for cartesians"""

    def get_renorm_pure(self, arg0: int, arg1: int, arg2: int) -> float:
        """Get renormalization constant for solid harmonics"""

    def get_shell_size(self, arg0: int) -> int: ...
    def print_atomic_info(self) -> None: ...
    def print_basis_info(self) -> None: ...
    def renormalize_contr(
        self, arg0: int, arg1: list[int], arg2: float
    ) -> None: ...
    def set_dummy_atoms(self, arg0: int) -> None: ...
    def to_hdf5(self, group: h5py._hl.group.Group) -> None:
        """Method implementing Basis.to_hdf5 dumper

        Args:
            group (h5py.Group): group in which Basis object information will be dumped

        """

    @property
    def alpha(self) -> list[float]:
        """The exponents for each primitive"""

    @alpha.setter
    def alpha(self, arg0: list[float]) -> None: ...
    @property
    def atom(self) -> list[int]:
        """The atomic numbers of each atom"""

    @atom.setter
    def atom(self, arg0: list[int]) -> None: ...
    @property
    def basisdir(self) -> str:
        """Get the directory of basis set library (if present)"""

    @property
    def basisname(self) -> str:
        """Get the name of the AO basis set (if specified)"""

    @property
    def contraction(self) -> list[float]:
        """The contraction coefficients for each primitive"""

    @contraction.setter
    def contraction(self, arg0: list[float]) -> None: ...
    @property
    def coordinates(self) -> list[list[float]]:
        """Get the coordinates"""

    @coordinates.setter
    def coordinates(self, arg0: list[list[float]]) -> None: ...
    @property
    def molfile(self) -> str:
        """Get the name of the coordinate xyz file (if present)"""

    @property
    def nbasis(self) -> int:
        """The total number of basis functions"""

    @nbasis.setter
    def nbasis(self, arg0: int) -> None: ...
    @property
    def ncenter(self) -> int:
        """The total number of atomic centers"""

    @ncenter.setter
    def ncenter(self, arg0: int) -> None: ...
    @property
    def nprim(self) -> list[int]:
        """The number of primitives per contraction"""

    @nprim.setter
    def nprim(self, arg0: list[int]) -> None: ...
    @property
    def nshell(self) -> int:
        """The total number of shells"""

    @nshell.setter
    def nshell(self, arg0: int) -> None: ...
    @property
    def shell2atom(self) -> list[int]:
        """The shell to atom list"""

    @shell2atom.setter
    def shell2atom(self, arg0: list[int]) -> None: ...
    @property
    def shell_types(self) -> list[int]:
        """The shell type per contraction"""

    @shell_types.setter
    def shell_types(self, arg0: list[int]) -> None: ...

class ExternalCharges:
    coordinates: list[list[float]]
    @typing.overload
    def __init__(
        self,
        arg0: list[float],
        arg1: list[float],
        arg2: list[float],
        arg3: list[float],
    ) -> None: ...
    @typing.overload
    def __init__(self, arg0: ExternalCharges) -> None: ...
    @property
    def charges(self) -> list[float]:
        """Charge values"""

    @charges.setter
    def charges(self, arg0: list[float]) -> None: ...

class StaticEmbedding:
    @typing.overload
    def __init__(
        self,
        arg0: list[float],
        arg1: list[float],
        arg2: list[float],
        arg3: list[float],
        arg4: list[float],
    ) -> None: ...
    @typing.overload
    def __init__(self, arg0: StaticEmbedding) -> None: ...
    def compute_static_embedding_cpp(
        self, arg0: Basis, arg1: Basis, arg2: NDArray[np.float64]
    ) -> None:
        """A function to calculate staticembedding taken from an external file"""

from . import basis, contract, ints, sapt_core, transform  # noqa

__all__ = [
    "Basis",
    "ExternalCharges",
    "StaticEmbedding",
    "basis",
    "contract",
    "ints",
    "sapt_core",
    "transform",
]
